/*
 * Stellarium Telescope Control Plug-in
 * 
 * Copyright (C) 2006 Johannes Gajdosik
 * Copyright (C) 2009 Bogdan Marinov
 * 
 * This module was originally written by Johannes Gajdosik in 2006
 * as a core module of Stellarium. In 2009 it was significantly extended with
 * GUI features and later split as an external plug-in module by Bogdan Marinov.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA  02110-1335, USA.
 */

#ifndef TELESCOPECLIENT_HPP
#define TELESCOPECLIENT_HPP

#include <QHostAddress>
#include <QHostInfo>
#include <QList>
#include <QString>
#include <QTcpSocket>
#include <QObject>

#include "StelApp.hpp"
#include "StelObject.hpp"
#include "common/InterpolatedPosition.hpp"

class StelCore;

qint64 getNow(void);

enum Equinox {
	EquinoxJ2000,
	EquinoxJNow
};

//! An abstract base class that should never be used directly, only inherited.
//! This class used to be called Telescope, but it has been renamed
//! to TelescopeClient in order to resolve a compiler/linker conflict
//! with the identically named Telescope class in Stellarium's main code.
class TelescopeClient : public QObject, public StelObject
{
	Q_OBJECT
public:
	static const QString TELESCOPECLIENT_TYPE;
	static TelescopeClient *create(const QString &url);
	virtual ~TelescopeClient(void) Q_DECL_OVERRIDE {}
	
	// Method inherited from StelObject
	QString getEnglishName(void) const Q_DECL_OVERRIDE {return name;}
	QString getNameI18n(void) const Q_DECL_OVERRIDE {return nameI18n;}
	Vec3f getInfoColor(void) const Q_DECL_OVERRIDE
	{
		return Vec3f(1.f, 1.f, 1.f);
	}
	//! TelescopeClient supports the following InfoStringGroup flags:
	//! - Name
	//! - RaDecJ2000
	//! - RaDec
	//! - PlainText
	//! @param core the StelCore object
	//! @param flags a set of InfoStringGroup items to include in the return value.
	//! @return a QString containing an HMTL encoded description of the Telescope.
	QString getInfoString(const StelCore* core, const InfoStringGroup& flags) const Q_DECL_OVERRIDE;
	QString getType(void) const Q_DECL_OVERRIDE {return TELESCOPECLIENT_TYPE;}
	QString getID() const Q_DECL_OVERRIDE {return name;}
	virtual double getAngularSize(const StelCore*) const Q_DECL_OVERRIDE {Q_ASSERT(0); return 0;}	// TODO
		
	// Methods specific to telescope
	virtual void telescopeGoto(const Vec3d &j2000Pos, StelObjectP selectObject) = 0;
	virtual void telescopeSync(const Vec3d &j2000Pos, StelObjectP selectObject) = 0;
	virtual void telescopeAbortSlew() {}

	//!
	//! \brief move
	//! \param angle [0,360)
	//! \param speed [0,1]
	//!
	virtual void move(double angle, double speed);
	virtual bool isConnected(void) const = 0;
	virtual bool hasKnownPosition(void) const = 0;
	void addOcular(double fov) {if (fov>=0.0) oculars.push_back(fov);}
	const QList<double> &getOculars(void) const {return oculars;}
	
	virtual bool prepareCommunication() {return false;}
	virtual void performCommunication() {}

	virtual QWidget* createControlWidget(QSharedPointer<TelescopeClient> telescope, QWidget* parent = Q_NULLPTR) const { Q_UNUSED(telescope) Q_UNUSED(parent) return Q_NULLPTR; }

protected:
	TelescopeClient(const QString &name);
	QString nameI18n;
	const QString name;

	virtual QString getTelescopeInfoString(const StelCore* core, const InfoStringGroup& flags) const
	{
		Q_UNUSED(core)
		Q_UNUSED(flags)
		return QString();
	}
private:
	virtual bool isInitialized(void) const {return true;}
	float getSelectPriority(const StelCore* core) const Q_DECL_OVERRIDE {Q_UNUSED(core); return -10.f;}
private:
	QList<double> oculars; // fov of the oculars
};

//! Example Telescope class. A physical telescope does not exist.
//! This can be used as a starting point for implementing a derived
//! Telescope class.
//! This class used to be called TelescopeDummy, but it had to be renamed
//! in order to resolve a compiler/linker conflict with the identically named
//! TelescopeDummy class in Stellarium's main code.
class TelescopeClientDummy : public TelescopeClient
{
public:
	TelescopeClientDummy(const QString &name, const QString &) : TelescopeClient(name)
	{
		desired_pos[0] = XYZ[0] = 1.0;
		desired_pos[1] = XYZ[1] = 0.0;
		desired_pos[2] = XYZ[2] = 0.0;
	}
	~TelescopeClientDummy(void) Q_DECL_OVERRIDE {}
	bool isConnected(void) const Q_DECL_OVERRIDE
	{
		return true;
	}
	bool prepareCommunication(void) Q_DECL_OVERRIDE
	{
		XYZ = XYZ * 31.0 + desired_pos;
		const double lq = XYZ.lengthSquared();
		if (lq > 0.0)
			XYZ *= (1.0/std::sqrt(lq));
		else
			XYZ = desired_pos;
		return true;
	}
	void telescopeGoto(const Vec3d &j2000Pos, StelObjectP selectObject) Q_DECL_OVERRIDE
	{
		Q_UNUSED(selectObject)
		desired_pos = j2000Pos;
		desired_pos.normalize();
	}
	void telescopeSync(const Vec3d &j2000Pos, StelObjectP selectObject) Q_DECL_OVERRIDE
	{
		Q_UNUSED(selectObject)
		Q_UNUSED(j2000Pos)
	}
	bool hasKnownPosition(void) const Q_DECL_OVERRIDE
	{
		return true;
	}
	Vec3d getJ2000EquatorialPos(const StelCore*) const Q_DECL_OVERRIDE
	{
		return XYZ;
	}
	
private:
	Vec3d XYZ; // j2000 position
	Vec3d desired_pos;
};

//! This TelescopeClient class can control a telescope by communicating
//! to a server process ("telescope server") via 
//! the "Stellarium telescope control protocol" over TCP/IP.
//! The "Stellarium telescope control protocol" is specified in a separate
//! document along with the telescope server software.
class TelescopeTCP : public TelescopeClient
{
	Q_OBJECT
public:
	TelescopeTCP(const QString &name, const QString &params, Equinox eq = EquinoxJ2000);
	~TelescopeTCP(void) Q_DECL_OVERRIDE
	{
		hangup();
	}
	bool isConnected(void) const Q_DECL_OVERRIDE
	{
		//return (tcpSocket->isValid() && !wait_for_connection_establishment);
		return (tcpSocket->state() == QAbstractSocket::ConnectedState);
	}
	
private:
	Vec3d getJ2000EquatorialPos(const StelCore* core=Q_NULLPTR) const Q_DECL_OVERRIDE;
	bool prepareCommunication() Q_DECL_OVERRIDE;
	void performCommunication() Q_DECL_OVERRIDE;
	void telescopeGoto(const Vec3d &j2000Pos, StelObjectP selectObject) Q_DECL_OVERRIDE;
	void telescopeSync(const Vec3d &j2000Pos, StelObjectP selectObject) Q_DECL_OVERRIDE;
	bool isInitialized(void) const Q_DECL_OVERRIDE
	{
		return (!address.isNull());
	}
	void performReading(void);
	void performWriting(void);
	
private:
	void hangup(void);
	QHostAddress address;
	quint16 port;
	QTcpSocket * tcpSocket;
	bool wait_for_connection_establishment;
	qint64 end_of_timeout;
	char readBuffer[120];
	char *readBufferEnd;
	char writeBuffer[120];
	char *writeBufferEnd;
	int time_delay;

	InterpolatedPosition interpolatedPosition;
	virtual bool hasKnownPosition(void) const Q_DECL_OVERRIDE
	{
		return interpolatedPosition.isKnown();
	}

	Equinox equinox;
	
private slots:
	void socketConnected(void);
	void socketFailed(QAbstractSocket::SocketError socketError);
};

#endif // TELESCOPECLIENT_HPP
